#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json
from typing import Dict, Any


class Entity:

    def __init__(self, client: 'AragornApiClient', metadata: Dict[str, Any]):
        self._client = client
        self._metadata = metadata

    @property
    def id(self) -> str:
        return self._metadata['id']

    def to_json(self, indent='    ', ensure_ascii: bool = False) -> str:
        return json.dumps(self._metadata, ensure_ascii=ensure_ascii, indent=indent)

    def __getitem__(self, key):
        return self._metadata[key]

    def __repr__(self):
        return repr(self._metadata)



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, Iterable
from pywaclient.endpoints import Endpoint


class CategoryEndpoint(Endpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client)
        self.path = 'category'
        self.path_articles = 'category/articles'

    def get(self, identifier: str, granularity: int) -> Dict:
        """Get the metadata of a specific category.

        :param identifier:  The identifier of the category.
        :param granularity: The granularity is between -1 and 2. Determines the amount of information returned by the
                            api.
        :return:            Category metadata
        """
        return self._get_request(self.path, {'id': identifier, 'granularity': str(granularity)})

    def put(self, content: Dict[str, Any]) -> Dict:
        """Create a new category. Required parameters for the creation of this entity are "title" and "world".

        :param content:     The content to create the category.
        :return:            Category metadata
        """
        return self._put_request(self.path, content)

    def patch(self, identifier: str, content: Dict[str, Any]) -> Dict[str, Any]:
        """
        Update (patch) the content of a category.

        {
            "state":"private",
            "bookcover":{
                "id":{{imageId}}
            }
        }

        :param identifier:  The id of the category to be updated.
        :param content:     All the fields that should be updated.
        :return:            Metadata of the updated category.
        """
        return self._patch_request(self.path, {'id': identifier}, content)

    def delete(self, identifier: str):
        """
        Delete a category.

        :param identifier:  The id of the category to be deleted.
        """
        self._delete_request(self.path, {'id': identifier})

    def articles(self, world_id: str, category_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[Dict[str, Any]]:
        """
        List all articles by a category given, filtered with a limit of entities shown and an offset.

        :param world_id:    The id of the world the articles should be returned from.
        :param category_id: The id of the category to return the articles from. To get articles without a category set
                            this value to -1. This does not return articles that have a parent article instead of a
                            category.
        :param complete     Ignore limit and offset and return all the articles as an iterable. Will fetch a new batch
                            every 50 articles.
        :param limit:       Determines how many articles are returned. Value between 1 and 50.
        :param offset:      Determines the offset at which articles are returned. Has to be a positive integer.
        :return:
        """
        if complete:
            return self._scroll_collection(self.path_articles, {'id': world_id}, 'entities', 'category', category_id)
        return self._post_request(self.path_articles,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset, 'category': {'id': category_id}})['entities']



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import logging
from typing import Optional, Dict, Iterable, List

from pywaclient.exceptions import AccessForbidden, InternalServerException
from pywaclient.models.entity import Entity
from pywaclient.models.genre import Genre
from pywaclient.models.article import Article
from pywaclient.models.category import Category
from pywaclient.models.block import Block
from pywaclient.models.secret import Secret


class World(Entity):

    def __init__(self, client: 'AragornApiClient', metadata: Dict[str, str]):
        super().__init__(client, metadata)

    @property
    def name(self) -> str:
        """Get the name of the world"""
        return self._metadata['name']

    @name.setter
    def name(self, name: str, update: bool = True):
        self._metadata['name'] = name
        if update:
            self._client.world.patch(self.id, {'name': name})

    @property
    def url(self) -> str:
        """Absolute url to the worlds homepage."""
        return self._metadata['url']

    @property
    def display_css(self) -> Optional[str]:
        """Property to get the world CSS if present.

        :return: Returns the CSS or None
        """
        if 'display_css' in self._metadata and self._metadata['display_css']:
            return self._metadata['display_css']
        else:
            return None

    @display_css.setter
    def display_css(self, css: str, update: bool = True):
        self._metadata['display_css'] = css
        if update:
            self._client.world.patch(self.id, {'display_css': css})

    @property
    def locale(self) -> str:
        """The language setting of this world as code (e.g. en, de, es etc)."""
        return self._metadata['locale']

    @property
    def genres(self) -> List[Genre]:
        """The genres of this world."""
        if 'genres' in self._metadata:
            for genre in self._metadata['genres']:
                yield Genre(genre)
        else:
            return []

    def categories(self) -> Iterable['Category']:
        for category in self._client.world.categories(self.id):
            if category is not None:
                yield Category(self._client, category)

    def articles(self) -> Iterable[Article]:
        for article in self._client.world.articles(self.id):
            try:
                article = self._client.article.get(article['id'], return_error=False)
                if article is not None:
                    yield Article(self._client, article)
            except AccessForbidden as err:
                logging.error(f"Access Forbidden: {err}.")
                yield None

    def blocks(self) -> Iterable[Block]:
        for block in self._client.world.blocks(self.id):
            try:
                block = self._client.block.get(block['id'], return_error=False)
                if block is not None:
                    yield Block(self._client, block)
            except InternalServerException as err:
                logging.error(f"Internal Server Error: {err}")

    def secrets(self) -> Iterable[Secret]:
        for secret in self._client.world.secrets(self.id):
            try:
                secret = self._client.secret.get(secret['id'], return_error=False)
                if secret is not None:
                    yield Secret(self._client, secret)
            except InternalServerException as err:
                logging.error(f"Internal Server Error: {err}")


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, Iterable, Union

from pywaclient.models.entity import Entity


class ManuscriptPart(Entity):

    def __init__(self, client: 'AragornApiClient', metadata: Dict[str, Any]):
        """

        :param client:
        :param metadata:
        """
        super().__init__(client, metadata)

    @property
    def type(self) -> str:
        """The type of this part. It is one value of folder, text and image."""
        return self._metadata['type']

    @property
    def title(self) -> str:
        """The title of the part."""
        return self._metadata['title'] if self._metadata['title'] is not None else ''


    @property
    def content(self) -> str:
        """The content of this part.

        - Folder: None
        - Text: Text
        - Image: Caption Text
        """
        return self._metadata['content'] if self._metadata['content'] is not None else ''

    @property
    def has_image(self) -> bool:
        """Checks if an image is present or not."""
        return self.type == 'image' and 'image' in self._metadata

    @property
    def image_url(self) -> Union[str, None]:
        """The url of the image if this is an image part."""
        if self.has_image:
            return self._metadata['image']['url']
        else:
            return None

    @property
    def image_id(self) -> Union[int, None]:
        """The id of the image if this is an image part."""
        if self.has_image:
            return self._metadata['image']['id']
        else:
            return None

    def children(self) -> Iterable['ManuscriptPart']:
        if 'children' in self._metadata:
            for key in self._metadata['children']:
                yield ManuscriptPart(self._client, self._metadata['children'][key])

    def to_html(self) -> str:
        if self.type == 'image':
            if self.has_image:
                return f'<img src="{self.image_url}" alt="{self.content}">\n'
            else:
                return ''
        elif self.type == 'text':
            return self.content
        else:
            return f'<h1>{self.title}</h1>\n'


class ManuscriptExport(Entity):

    def __init__(self, client: 'AragornApiClient', metadata: Dict[str, Any], manuscript: 'Manuscript' = None):
        """

        :param client:
        :param metadata:
        """
        super().__init__(client, metadata)
        self._manuscript = manuscript

    @property
    def manuscript(self) -> 'Manuscript':
        if self._manuscript is None:
            return Manuscript.from_id(self._client, self._metadata['manuscript']['id'])
        else:
            return self._manuscript

    @property
    def version(self) -> 'ManuscriptVersion':
        return ManuscriptVersion(self._client, self._metadata['version'])

    @property
    def parts(self) -> Iterable['ManuscriptPart']:
        if 'parts' in self._metadata:
            for key in self._metadata['parts']:
                yield ManuscriptPart(self._client, self._metadata['parts'][key])


class ManuscriptVersion(Entity):

    def __init__(self, client: 'AragornApiClient', metadata: Dict[str, Any]):
        super().__init__(client, metadata)

    @property
    def title(self) -> str:
        return self._metadata['title']

    @property
    def state(self) -> str:
        """"""
        return self._metadata['state']


class Manuscript(Entity):

    @classmethod
    def from_id(cls, client: 'AragornApiClient', identifier: str):
        return Manuscript(client, client.manuscript.get(identifier))

    def __init__(self, client: 'AragornApiClient', metadata: Dict[str, Any]):
        """

        :param client:
        :param metadata:
        .:param author:
        """
        super().__init__(client, metadata)

    @property
    def title(self) -> str:
        """Title of the manuscript.

        :return: manuscript title
        """
        return self._metadata['title']

    @property
    def status(self) -> str:
        """The status of this manuscript (completed / ongoing).

        :return: status
        """
        return self._metadata['status']

    @property
    def author(self) -> 'User':
        """The author of this manuscript.

        :return: User
        """
        # this import is here because of circular imports.
        from pywaclient.models.user import User
        return User(self._client, self._metadata['author']['id'])

    @property
    def active_version(self) -> ManuscriptVersion:
        """Fetches the active version of this manuscript.

        :return: ManuscriptVersion
        """
        return ManuscriptVersion(self._client, self._metadata['active_version'][0])

    @property
    def active_export(self) -> ManuscriptExport:
        """Fetches the export of the active version of this manuscript.

        :return: ManuscriptExport
        """
        return ManuscriptExport(self._client, self._client.manuscript.export_version(self.active_version.id), self)



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any

from pywaclient.endpoints import Endpoint


class MapMarkerEndpoint(Endpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client)
        self.path = 'marker'

    def get(self, identifier: str, granularity: int) -> Dict[str, Any]:
        """Get the metadata of a specific map marker.

        :param identifier:  The identifier of the map marker.
        :param granularity: The granularity is between -1 and 2. Determines the amount of information returned by the
                            api.
        :return:            Map marker metadata
        """
        return self._get_request(self.path, {'id': identifier, 'granularity': str(granularity)})

    def put(self, content: Dict[str, Any]) -> Dict[str, Any]:
        """
        Create a new map marker. Required parameters for the creation of this entity are "title" and "world.id".

        {
            "title": "Town Map marker",
            "world": {
                "id":"{{worldId}}"
            }
        }

        :param content:     Metadata of the new map marker with at least he required fields.
        :return:            Metadata of the created map marker with the id.
        """
        return self._put_request(self.path, content)

    def patch(self, identifier: str, content: Dict[str, Any]) -> Dict[str, Any]:
        """
        Update (patch) the content of a map marker.

        :param identifier:  The id of the map marker to be updated.
        :param content:     All the fields that should be updated.
        :return:            Metadata of the map marker.
        """
        return self._patch_request(self.path, {'id': identifier}, content)

    def delete(self, identifier: str):
        """
        Delete a map marker.

        :param identifier:  The id of the map marker to be deleted.
        """
        self._delete_request(self.path, {'id': identifier})


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import world_id, client

if __name__ == '__main__':
    block = client.block.get('678589', 2)
    print(block)

    test_block_1 = client.block.put(
        {
            "title": "New Block",
            "RPGSRD": {
                "id": "273"
            },
            'template': {
                'id': '27'
            }
        }
    )
    test_block_2 = client.block.put(
        {
            "title": "New Block",
            "RPGSRD": {
                "id": "273"
            }
        }
    )
    response_patch_block_2 = client.block.patch(
        test_block_2['id'],
        {
            'excerpt': 'This is an excerpt for an block.'
        }
    )

    full_test_block_2 = client.block.get(
        test_block_2['id'],
        2
    )

    assert full_test_block_2['excerpt'] == 'This is an excerpt for an block.'

    client.block.delete(test_block_1['id'])
    client.block.delete(test_block_2['id'])



from typing import Dict


class WorldAnvilClientException(Exception):
    """Base exception of the library."""


class NoParentCategoryException(WorldAnvilClientException):
    """When a category does not have a parent category this exception is thrown."""


class WorldAnvilServerException(WorldAnvilClientException):
    """Exceptions returned by the server for requests made."""


class ConnectionException(WorldAnvilClientException):
    """Was unable to connect to World Anvil for some reason."""


class UnexpectedStatusException(WorldAnvilServerException):
    """An unexpected status exception occurred."""

    def __init__(self, status: int, message: str, path: str):
        self.status = status
        self.message = message
        self.path = path


class InternalServerException(WorldAnvilServerException):
    """Internal Server Error in World Anvil Response."""

    def __init__(self, path: str):
        self.status = 500
        self.message = "Unable to process the request."
        self.path = path


class AccessForbidden(WorldAnvilServerException):
    """The user does not have permissions for the requested resources."""

    def __init__(self, path: str):
        self.status = 403
        self.message = "No permission to access the requested resource."
        self.path = path


class ResourceNotFound(WorldAnvilServerException):
    """The requested resource does not exist or was moved."""

    def __init__(self, path: str):
        self.status = 404
        self.message = "Request resource was not found."
        self.path = path


class UnprocessableDataProvided(WorldAnvilServerException):
    """The request could not be processed."""

    def __init__(self, path: str, data: Dict[str, str]):
        self.status = 422
        self.path = path
        if 'status' in data:
            self.message = data['status']
            self.error_summary = data['error']['summary']
            self.error_tracestack = data['error']['traceStack']
        else:
            self.message = data['error']
            self.error_tracestack = data['trace']


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Dict, Any, Iterable

from pywaclient.endpoints import Endpoint


class WorldEndpoint(Endpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client)
        self.path = 'world'
        self.path_categories = 'world/categories'
        self.path_block_folders = 'world/blockfolders'
        self.path_variable_collections = 'world/variablecollections'
        self.path_subscriber_groups = 'world/subscribergroups'
        self.path_secrets = 'world/secrets'
        self.path_manuscripts = 'world/manuscripts'
        self.path_maps = 'world/maps'
        self.path_images = 'world/images'
        self.path_histories = 'world/histories'
        self.path_timelines = 'world/timelines'
        self.path_chronicles = 'world/chronicles'
        self.path_canvases = 'world/canvases'

    def get(self, identifier: str, granularity: int) -> Dict[str, Any]:
        """Get the metadata of a specific world.

        :param identifier:  Identifier of the world.
        :param granularity: The level of details of the returned metadata. Is a number from -1 to 2.
        :return:            World metadata.
        """
        assert -1 <= granularity <= 2
        return self._get_request(self.path, {'id': identifier, 'granularity': str(granularity)})

    def put(self, content: Dict[str, Any]) -> Dict[str, Any]:
        """
        Create a new world. Required parameters for the creation of this entity are "title".

        {
            "title": "A bright new World"
        }

        :param content:     The fields and their values with which to create the new world.
        :return:            The entity of the new world.
        """
        return self._put_request(self.path, content)

    def patch(self, identifier: str, content: Dict[str, Any]) -> Dict[str, Any]:
        """Update (patch) the world configuration and fields.

        :param identifier:  The identifier of the world.
        :param content:     The fields and values which should be updated.
        :return:            The world entity.
        """
        return self._patch_request(self.path, {'id': identifier}, content)

    def delete(self, identifier: str):
        """
        Delete a world.

        :param identifier:  The id of the world to be deleted.
        """
        self._delete_request(self.path, {'id': identifier})

    def categories(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all categories by a world given, filtered with a limit of entities shown and an offset.

        :param world_id:    The id of the world the folders should be returned from.
        :param complete:    Return the entire list of entities as a list. Only loads 50 categories at a time.
        :param limit:       Determines how many folders are returned. Value between 1 and 50.
        :param offset:      Determines the offset at which folders are returned. Has to be a positive integer.
        :return:
        """
        if complete:
            return self._scroll_collection(self.path_categories, {'id': world_id}, 'entities')
        return self._post_request(self.path_categories,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def statblock_folders(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all statblock folders of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of statblock folder entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_block_folders, {'id': world_id}, 'entities')
        return self._post_request(self.path_block_folders,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def variable_collections(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all variable collections of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of variable collection entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_variable_collections, {'id': world_id}, 'entities')
        return self._post_request(self.path_variable_collections,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def subscriber_groups(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all subscriber groups of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of subscriber group entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_subscriber_groups, {'id': world_id}, 'entities')
        return self._post_request(self.path_subscriber_groups,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def secrets(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all secrets of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of secret entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_secrets, {'id': world_id}, 'entities')
        return self._post_request(self.path_secrets,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def manuscripts(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all manuscripts of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of manuscript entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_manuscripts, {'id': world_id}, 'entities')
        return self._post_request(self.path_manuscripts,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def maps(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all maps of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of map entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_maps, {'id': world_id}, 'entities')
        return self._post_request(self.path_maps,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def images(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all images of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of image entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_images, {'id': world_id}, 'entities')
        return self._post_request(self.path_images,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def histories(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all histories of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of history entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_histories, {'id': world_id}, 'entities')
        return self._post_request(self.path_histories,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def timelines(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all timelines of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of timeline entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_timelines, {'id': world_id}, 'entities')
        return self._post_request(self.path_timelines,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def chronicles(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all chronicles of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of chronicle entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_chronicles, {'id': world_id}, 'entities')
        return self._post_request(self.path_chronicles,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']

    def canvases(self, world_id: str, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[
        Dict[str, Any]]:
        """
        List all canvases of a specific world.

        :param world_id:    The id of the world
        :param complete:    Return the entire list of entities as an iterable. This ignores limit and offset when
                            set to true.
        :param limit:       The number of entries returned per request. 1 <= limit <= 50
        :param offset:      The offset in the list of entries to return from. offset >=
        :return:            An iterable of canvas entities.
        """
        assert 1 <= limit <= 50
        assert offset >= 0
        if complete:
            return self._scroll_collection(self.path_canvases, {'id': world_id}, 'entities')
        return self._post_request(self.path_canvases,
                                  {'id': world_id},
                                  {'limit': limit, 'offset': offset})['entities']



#    Copyright 2020 - present Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os

from api import BoromirApiClient
from models.user import User

world_id = 'daae0a12-f3c3-4978-b571-b5313e3c1741'
client = BoromirApiClient(
    name='TEST APPLICATION',
    url='https://gitlab.com/SoulLink/world-anvil-api-client',
    application_key=os.environ.get('WA_APPLICATION_KEY'),
    authentication_token=os.environ.get('WA_AUTH_TOKEN'),
    version='0.1.0'
)
user = User(client, client.user.identity()['id'])


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import client

if __name__ == '__main__':
    notebook = client.notebook.get('5b794d91-2730-4b76-aa21-dddbabdc1c58', 2)
    print(notebook)

    for s in client.notebook.note_sections(notebook['id']):
        print(s)
        for n in client.note_section.notes(s['id']):
            print(n)

    for s in notebook['notesections']:
        section = client.note_section.get(s['id'], 2)
        print(section)

        client.note.put(
            {'title': 'A new note', 'content': 'Some content', 'notesection': {'id': section['id']}}
        )

    # client.note_section.put(
    #     {
    #         'title': 'A new section',
    #         'notebook': {
    #             'id': notebook['id']
    #         }
    #     }
    #
    # )

    # value = client.notebook.put(
    #     {
    #         'title': "A New Notebook",
    #         'owner': {
    #             'id': '42eb1d6a-021b-49b4-bbbb-f7ddf6b135a4'
    #         }
    #     }
    # )
    # print(value)


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json
from typing import Dict, Any, Iterable, Optional

import requests
from requests import RequestException

from pywaclient.exceptions import AccessForbidden, ResourceNotFound, InternalServerException, \
    UnexpectedStatusException, ConnectionException, UnprocessableDataProvided


def _parse_response(path, response) -> Dict[str, Any]:
    if response.ok:
        data = response.json()
        if data['success']:
            return data
        else:
            raise UnexpectedStatusException(response.status_code, data['error'], path)
    elif response.status_code == 403:
        raise AccessForbidden(path)
    elif response.status_code == 404:
        raise ResourceNotFound(path)
    elif response.status_code == 422:
        raise UnprocessableDataProvided(path, response.json())
    elif response.status_code == 500:
        raise InternalServerException(path)
    else:
        raise UnexpectedStatusException(response.status_code, response.reason, path)


class Endpoint:

    def __init__(self, client: 'AragornApiClient'):
        self.client = client

    def _get_request(self, path: str, params: Dict[str, str]) -> Dict[str, Any]:
        try:
            response = requests.get(self.client.base_url + path, params=params, headers=self.client.headers)
            return _parse_response(path, response)
        except RequestException as err:
            raise ConnectionException(str(err))

    def _put_request(self, path: str, content: Dict[str, Any]) -> Dict[str, Any]:
        try:
            response = requests.put(self.client.base_url + path, json=content, headers=self.client.headers_post)
            return _parse_response(path, response)
        except RequestException as err:
            raise ConnectionException(str(err))

    def _patch_request(self, path: str, params: Dict[str, str], content: Dict[str, Any]) -> Dict[str, Any]:
        try:
            response = requests.patch(self.client.base_url + path, params=params, json=content,
                                      headers=self.client.headers_post)
            return _parse_response(path, response)
        except RequestException as err:
            raise ConnectionException(str(err))

    def _post_request(self, path: str, params: Dict[str, str], content: Dict[str, Any]) -> Dict[str, Any]:
        try:
            response = requests.post(self.client.base_url + path, params=params, json=content,
                                     headers=self.client.headers_post)
            return _parse_response(path, response)
        except RequestException as err:
            raise ConnectionException(str(err))

    def _delete_request(self, path: str, params: Dict[str, str]):
        try:
            response = requests.delete(self.client.base_url + path,
                                       params=params,
                                       headers=self.client.headers)
            return _parse_response(path, response)
        except RequestException as err:
            raise ConnectionException(str(err))

    def _scroll_collection(self, path: str, params: Dict[str, str], collection_tag: str, parentName: str = '', parentId: str = '') -> Iterable[Dict[str, Any]]:
        limit = 50
        offset = 0
        if parentName != '':
            content = {'limit': limit, 'offset': offset, parentName: {'id': parentId}}
        else:
            content = {'limit': limit, 'offset': offset}
        collection = self._post_request(path, params, content)
        if collection['success']:
            items = collection[collection_tag]
            while len(items) > 0:
                for item in items:
                    yield item
                offset = offset + limit
                content['offset'] = offset
                collection = self._post_request(path, params, content)
                items = collection[collection_tag]

    @staticmethod
    def _download_binary(url: str, chunk_size: int = 8192) -> Iterable[bytes]:
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=chunk_size):
                yield chunk



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from pywaclient.endpoints.maps import MapEndpoint
from pywaclient.endpoints.rpg_system import RpgSystemEndpoint
from pywaclient.endpoints.variables import VariableEndpoint
from pywaclient.endpoints.variable_collections import VariableCollectionEndpoint
from pywaclient.endpoints.secrets import SecretEndpoint
from pywaclient.endpoints.categories import CategoryEndpoint
from pywaclient.endpoints.articles import ArticleEndpoint
from pywaclient.endpoints.blocks import BlockEndpoint
from pywaclient.endpoints.images import ImageEndpoint
from pywaclient.endpoints.manuscripts import ManuscriptEndpoint
from pywaclient.endpoints.users import UserEndpoint
from pywaclient.endpoints.worlds import WorldEndpoint
from pywaclient.endpoints.subscriber_groups import SubscriberGroupEndpoint
from pywaclient.endpoints.histories import HistoryEndpoint
from pywaclient.endpoints.timelines import TimelineEndpoint
from pywaclient.endpoints.chronicles import ChronicleEndpoint
from pywaclient.endpoints.canvas import CanvasEndpoint
from pywaclient.endpoints.block_templates import BlockTemplateEndpoint
from pywaclient.endpoints.block_template_parts import BlockTemplatePartEndpoint
from pywaclient.endpoints.notes import NoteEndpoint
from pywaclient.endpoints.notebooks import NotebookEndpoint
from pywaclient.endpoints.notesections import NoteSectionEndpoint


class BoromirApiClient:

    def __init__(self,
                 name: str,
                 url: str,
                 version: str,
                 application_key: str,
                 authentication_token: str,
                 ):
        self.headers = {
            'x-auth-token': authentication_token,
            'x-application-key': application_key,
            'Accept': 'application/json',
            'User-Agent': f'{name} ({url}, {version})'
        }
        self.headers_post = self.headers.copy()
        self.headers_post['Content-type'] = 'application/json'
        self.base_url = 'https://www.worldanvil.com/api/external/boromir/'
        self.block = BlockEndpoint(self)
        self.article = ArticleEndpoint(self)
        self.image = ImageEndpoint(self)
        self.manuscript = ManuscriptEndpoint(self)
        self.user = UserEndpoint(self)
        self.secret = SecretEndpoint(self)
        self.world = WorldEndpoint(self)
        self.category = CategoryEndpoint(self)
        self.variable_collection = VariableCollectionEndpoint(self)
        self.variable = VariableEndpoint(self)
        self.rpg_system = RpgSystemEndpoint(self)
        self.subscriber_group = SubscriberGroupEndpoint(self)
        self.map = MapEndpoint(self)
        self.history = HistoryEndpoint(self)
        self.timeline = TimelineEndpoint(self)
        self.canvas = CanvasEndpoint(self)
        self.chronicle = ChronicleEndpoint(self)
        self.block_template = BlockTemplateEndpoint(self)
        self.block_template_part = BlockTemplatePartEndpoint(self)
        self.note = NoteEndpoint(self)
        self.note_section = NoteSectionEndpoint(self)
        self.notebook = NotebookEndpoint(self)


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json

from init_client import world_id, client

if __name__ == '__main__':
    premade_group_id = '92e74dbb-3ebe-4a07-999f-4e8dfb77fb33'

    group = client.subscriber_group.get(premade_group_id, 2)
    print(group)

    client.subscriber_group.patch(group['id'],
                                  {
                                      'title': 'A new title',
                                      'description': 'A different description',
                                      'position': 100,
                                      'isHidden': False,
                                      'isAssignable': False,
                                      'tags': 'anew,tag,a,day',
                                      'isDefault': False
                                  }
                                  )

    groups = client.world.subscriber_groups(world_id)
    for s in groups:
        print(json.dumps(client.subscriber_group.get(s['id'], 2), indent='    ', ensure_ascii=False))
    test_subscriber_group_1 = client.subscriber_group.put(
        {
            'title': 'Test Subscriber Group Creation',
            'world': {
                'id': world_id}
        }
    )
    test_subscriber_group_2 = client.subscriber_group.put(
        {
            'title': 'Test Subscriber Group Creation 2',
            'world': {
                'id': world_id}
        }
    )
    response_patch_subscriber_group_2 = client.subscriber_group.patch(
        test_subscriber_group_2['id'],
        {
            'title': 'A different subscriber group'
        }
    )

    full_test_subscriber_group_2 = client.subscriber_group.get(
        test_subscriber_group_2['id'],
        2
    )
    print(full_test_subscriber_group_2)

    assert full_test_subscriber_group_2['title'] == 'A different subscriber group'

    client.subscriber_group.delete(test_subscriber_group_1['id'])
    client.subscriber_group.delete(test_subscriber_group_2['id'])



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, Iterable

from pywaclient.endpoints import Endpoint


class UserEndpoint(Endpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client)
        self.identity_path = 'identity'
        self.path = 'user'
        self.path_blocktemplates = 'user/blocktemplates'
        self.path_worlds = 'user/worlds'

    def identity(self):
        """Get the user that is associated with the authentication token provided."""
        return self._get_request(self.identity_path, {})

    def get(self, identifier: str, granularity: int) -> Dict[str, Any]:
        """Retrieve a user with their identifier.

        :param identifier:  The identifier of the user.
        :param granularity: A value between -1 and 2. Determines how much metadata is returned.
        :return:            User metadata.
        """
        return self._get_request(self.path, {'id': identifier, 'granularity': str(granularity)})

    def update(self, identifier: str, content: Dict[str, Any]) -> Dict[str, str]:
        """Update (patch) the content of a user.

        :param identifier:  The identifier of the user.
        :param content:     The fields to be updated.
        :return:            Basic user metadata.
        """
        return self._patch_request(self.path, {'id': identifier}, content)

    def block_templates(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Returns a list of all the block templates owned by this user.

        :param identifier:  The identifier of the user that the block templates should be returned from.
        :return:            An iterable of the block template entities.
        """
        return self._scroll_collection(self.path_blocktemplates, {'id': identifier}, 'entities')

    def worlds(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Returns a list of all worlds owned by this user.

        :param identifier:  The identifier of the user that the worlds should be returned from.
        :return:            An iterable of the world entities.
        """
        return self._scroll_collection(self.path_worlds, {'id': identifier}, 'owned')

    def authored_worlds(self, identifier: str) -> Iterable[Dict[str, Any]]:
        """Returns a list of all worlds that the user can author (co-owner, editor, writer).

        :param identifier:  The identifier of the user that the worlds should be returned from.
        :return:            An iterable of the world entities.
        """
        return self._scroll_collection(self.path_worlds, {'id': identifier}, 'author')



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, List, Union

from pywaclient.models.entity import Entity


class Article(Entity):

    def __init__(self, client: 'AragornApiClient', metadata: Dict[str, Any]):
        super().__init__(client, metadata)
        if 'fullRender' in self._metadata:
            self._full_render = self._metadata['fullRender']
            del self._metadata['fullRender']
        else:
            self._full_render = ""

    @property
    def title(self) -> str:
        return self._metadata['title']

    @property
    def template(self) -> str:
        return self._metadata['template']

    @property
    def is_draft(self) -> bool:
        return self._metadata['isDraft']

    @property
    def is_wip(self) -> bool:
        return self._metadata['isWip']

    @property
    def state(self) -> str:
        """State of the article: 'private' or 'public'."""
        return self._metadata['state']

    @property
    def wordcount(self) -> int:
        """Wordcount within the article."""
        return self._metadata['wordcount']

    @property
    def url(self) -> str:
        """Absolute url of the article."""
        if 'url' in self._metadata:
            return self._metadata['url']
        else:
            return ''

    @property
    def tags(self) -> List[str]:
        """A list of all the tags on this article."""
        tags = self._metadata['tags']
        if tags is None:
            return list()
        else:
            return tags.split(',')

    @property
    def last_update(self) -> str:
        """The date when the article was update the last time."""
        if 'lastUpdate' in self._metadata and 'date' in self._metadata['lastUpdate']:
            return self._metadata['lastUpdate']['date']
        elif 'updateDate' in self._metadata and 'date' in self._metadata['updateDate']:
            return self._metadata['updateDate']['date']
        else:
            return ''

    @property
    def creation_date(self) -> str:
        """The date when this article was created."""
        if 'creationDate' in self._metadata:
            value = self._metadata['creationDate']
            if value is None:
                return ''
            else:
                return self._metadata['creationDate']['date']
        else:
            return ''

    @property
    def notification_date(self) -> str:
        """The date when the last notification was sent out."""
        if 'notificationDate' in self._metadata:
            value = self._metadata['notificationDate']
            if value is None:
                return ''
            else:
                return self._metadata['notificationDate']['date']
        else:
            return ''

    @property
    def publication_date(self) -> str:
        """The date when the last notification was sent out."""
        if 'publicationDate' in self._metadata:
            value = self._metadata['publicationDate']
            if value is None:
                return ''
            else:
                return self._metadata['publicationDate']['date']
        else:
            return ''

    @property
    def has_category(self) -> bool:
        return 'category' in self._metadata

    @property
    def category_title(self) -> str:
        if 'category' in self._metadata:
            return self._metadata['category']['title']
        else:
            return ''

    @property
    def category_id(self) -> str:
        if 'category' in self._metadata:
            return self._metadata['category']['id']
        else:
            return ''

    @property
    def category(self) -> Union['Category', None]:
        if self.has_category:
            from pywaclient.models.category import Category
            return Category(self._client, self._client.category.get(self.category_id))
        else:
            return None

    @property
    def has_parent_article(self) -> bool:
        if 'relations' in self._metadata:
            return 'articleParent' in self._metadata['relations']
        else:
            return False

    @property
    def parent_article_id(self) -> str:
        """Returns the parent article id or an empty string if it is not defined."""
        if self.has_parent_article:
            try:
                return self._metadata['relations']['articleParent']['items']['id']
            except KeyError:
                return self._metadata['relations']['articleParent']['id']
        else:
            ''

    @property
    def parent_article_title(self) -> str:
        """Returns the parent article title or an empty string if it is not defined."""
        if self.has_parent_article:
            try:
                return self._metadata['relations']['articleParent']['items']['title']
            except KeyError:
                return self._metadata['relations']['articleParent']['title']
        else:
            ''

    def get_parent_article(self) -> Union['Article', None]:
        """Get the model of the parent article in full. This makes a  GET request to WorldAnvil."""
        if self.has_parent_article:
            identifier = self.parent_article_id
            if identifier == '':
                return None
            return Article(self._client, self._client.article.get(identifier))
        else:
            return None

    @property
    def author_id(self) -> str:
        return self._metadata['author']['id']

    @property
    def author(self) -> 'User':
        from pywaclient.models.user import User
        return User(self._client, self.author_id)

    @property
    def world_id(self) -> str:
        return self._metadata['world']['id']

    @property
    def world(self) -> 'World':
        from pywaclient.models.world import World
        return World(self._client, self._client.world.get(self.world_id))

    @property
    def full_render(self) -> str:
        return self._full_render

    @property
    def excerpt(self) -> str:
        if 'excerpt' in self._metadata:
            return self._metadata['excerpt']
        else:
            return ''

    def to_json(self, indent='    ', ensure_ascii: bool = False, include_full_render: bool = False) -> str:
        if include_full_render:
            self._metadata['fullRender'] = self._full_render
        return super().to_json(indent, ensure_ascii)



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any, Iterable

from pywaclient.endpoints import Endpoint


class RpgSystemEndpoint(Endpoint):

    def __init__(self, client: 'AragornApiClient'):
        super().__init__(client)
        self.path = 'rpgsystem'
        self.list_path = 'rpgsystems'

    def get(self, identifier: str, granularity: int) -> Dict[str, Any]:
        """Get the metadata of a specific rpg system.

        :param identifier:  The identifier of the article.
        :param granularity: The granularity is between -1 and 2. Determines the amount of information returned by the
                            api.
        :return:            RPG System metadata
        """
        return self._get_request(self.path, {'id': identifier, 'granularity': str(granularity)})

    def list(self, complete: bool = True, limit: int = 50, offset: int = 0) -> Iterable[Dict[str, Any]]:
        """
        List all RPG systems, filtered with a limit of entities shown and an offset.

        :param complete:    Ignore the limit and offsets and return all the entities.
        :param limit:       Determines how many RPG systems are returned. Value between 1 and 50.
        :param offset:      Determines the offset at which RPG systems are returned. Has to be a positive integer.
        :return:
        """
        if complete:
            return self._scroll_collection(self.list_path, {}, 'entities')
        assert offset >= 0
        assert 1 <= limit <= 50
        return self._post_request(self.list_path, {}, {'limit': limit, 'offset': offset})['entities']


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from init_client import world_id, client

if __name__ == '__main__':
    template = client.block_template.get('3381', 1)
    print(template)

    test_block_template_1 = client.block_template.put(
        {
            "title": "5e TEST Block Template",
            "formSchemaParser": "json",
            "formSchema": '{"version":"1","system":"fate","author":{"username":"SoulLink","email":"jonaswaeber@gmail.com"},"unique_reference":"average_nameless_npc-soullink-4","name":"Average Nameless NPC","display_template":"","description":"The weakest of all adversary. Exist to make the players look awesome!","fields":{"name":{"input":"string","label":"Name","placeholder":"The name of theAverage Nameless NPC","required":true},"character_aspect_1":{"input":"string","label":"Character Aspect 1","description":"A generic aspect that can be used against this character.","placeholder":""},"character_aspect_2":{"input":"string","label":"Character Aspect 2","description":"An additional character aspect.","placeholder":""},"average_1_skill_1":{"input":"select","label":"Average (+1) Skill 1","description":"Should support whatever this NPC is going to use to compete in a conflict, contest or challenge.","placeholder":"","options":{"None":"None","Athletics":"Athletics","Burglary":"Burglary","Contacts":"Contacts","Crafts":"Crafts","Deceive":"Deceive","Drive":"Drive","Empathy":"Empathy","Fight":"Fight","Investigate":"Investigate","Lore":"Lore","Notice":"Notice","Physique":"Physique","Provoke":"Provoke","Rapport":"Rapport","Resources":"Resources","Shoot":"Shoot","Stealth":"Stealth","Will":"Will"}},"average_1_skill_2":{"input":"select","label":"Average (+1) Skill 2","description":"An additional support skill.","placeholder":"","options":{"None":"None","Athletics":"Athletics","Burglary":"Burglary","Contacts":"Contacts","Crafts":"Crafts","Deceive":"Deceive","Drive":"Drive","Empathy":"Empathy","Fight":"Fight","Investigate":"Investigate","Lore":"Lore","Notice":"Notice","Physique":"Physique","Provoke":"Provoke","Rapport":"Rapport","Resources":"Resources","Shoot":"Shoot","Stealth":"Stealth","Will":"Will"}}}}',
            "RPGSRD": {
                "id": 1
            }
        }
    )
    test_block_template_2 = client.block_template.put(
        {
            'title': 'Test block_template Creation 2',
            'templateType': 'block_template',
            'world': {
                'id': world_id}
        }
    )
    response_patch_block_template_2 = client.block_template.patch(
        test_block_template_2['id'],
        {
            'excerpt': 'This is an excerpt for an block_template.'
        }
    )

    full_test_block_template_2 = client.block_template.get(
        test_block_template_2['id'],
        2
    )

    assert full_test_block_template_2['excerpt'] == 'This is an excerpt for an block_template.'

    client.block_template.delete(test_block_template_1['id'])
    client.block_template.delete(test_block_template_2['id'])

    block_template_with_a_lot_of_views = client.block_template.put(
        {
            'title': 'An block_template with a lot of views.',
            'templateType': 'block_template',
            'world': {
                'id': world_id
            }
        }
    )
    print(client.block_template.get(block_template_with_a_lot_of_views['id'], 2))



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from pywaclient.exceptions import UnprocessableDataProvided, InternalServerException
from init_client import client

if __name__ == '__main__':
    keys = set()
    for item in client.rpg_system.list():
        if item['id'] not in [274, 267, 14, 18, 288, 287, 501, 358, 366, 372]:
            try:
                system = client.rpg_system.get(item['id'], 2)
            except UnprocessableDataProvided as err:
                print(item['id'], item['title'])
                print(err.status)
                print(err.error_summary)
                print(err.error_tracestack)
            except InternalServerException as err:
                print('Internal Server Error', item['id'], item['title'])
        else:
            print(item['title'], item['id'])
        keys.update(system.keys())
        assert item['entityClass'] == 'RPGSRD'
    print(keys)

    {'styles', 'entityClass', 'blockTemplates', 'url', 'publisher', 'author', 'campaigns', 'slug', 'subscribergroups',
     'folderId', 'state', 'blocks', 'worlds', 'description', 'success', 'icon', 'isDraft', 'heroesEnabled', 'title',
     'id', 'weight', 'updateDate', 'copyright', 'isWip', 'characters', 'reference', 'logo', 'isEditable', 'tags'}


#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import logging
from typing import Optional, Iterable

from pywaclient.models.entity import Entity
from pywaclient.models.manuscript import Manuscript
from pywaclient.models.world import World


class User(Entity):

    def __init__(self, client: 'AragornApiClient', identifier: str):
        metadata = client.user.get(identifier, 0)
        super().__init__(client, metadata)

    @property
    def username(self) -> str:
        return self._metadata['username']

    def world(self, identifier: str) -> World:
        return World(self._client, self._client.world.get(identifier))

    def worlds(self) -> Iterable[World]:
        result = self._client.user.worlds(self.id)
        if result:
            for world in result:
                world = self._client.world.get(world['id'])
                if world is not None:
                    yield World(self._client, world)
                else:
                    logging.error("Could not retrieve a world because of an error in the API.")

    def manuscripts(self) -> Iterable[Manuscript]:
        result = self._client.user.manuscripts(self.id)
        if result:
            for manuscript in result:
                manuscript = self._client.manuscript.get(manuscript['id'])
                if manuscript is not None:
                    yield Manuscript(self._client, manuscript)
                else:
                    logging.error("Could not retrieve a manuscript because of an error in the API.")





#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any

from pywaclient.models.entity import Entity
from pywaclient.exceptions import NoParentCategoryException


class Category(Entity):

    def __init__(self, client: 'AragornApiClient', metadata: Dict[str, Any]):
        super().__init__(client, metadata)
        self._world = None

    @property
    def title(self) -> str:
        """Title of the category."""
        return self._metadata['title']

    @property
    def url(self) -> str:
        """url of the category."""
        return self._metadata['url']

    @property
    def state(self) -> str:
        """State of the category: public | private."""
        return self._metadata['state']

    @property
    def views(self) -> int:
        """Number of views on the category."""
        views = self._metadata['views']
        return views if views else 0

    @property
    def position(self) -> int:
        """Number of views on the category."""
        position = self._metadata['position']
        return position if position else 0

    @property
    def description(self) -> str:
        """The main description field of the category."""
        description = self._metadata['description']
        return description if description else ""

    @property
    def excerpt(self) -> str:
        """The main excerpt field of the category."""
        excerpt = self._metadata['excerpt']
        return excerpt if excerpt else ""

    @property
    def icon(self) -> str:
        """The main icon field of the category."""
        icon = self._metadata['icon']
        return icon if icon else ""

    @property
    def has_parent_category(self) -> bool:
        """Has this category a parent category?"""
        return 'parent' in self._metadata and self._metadata['parent'] is not None

    @property
    def parent_category(self) -> 'Category':
        """Get the parent category of this category.

        :returns: Parent category
        :raises NoParentCategoryException: No parent category present.
        """
        if self.has_parent_category:
            return Category(self._client, self._client.category.get(self._metadata['parent']['id']))
        else:
            raise NoParentCategoryException(f"The category {self.title} has no parent category.")

    @property
    def book_cover(self) -> str:
        """Returns the book cover url or an empty string."""
        if 'book_cover' in self._metadata:
            return self._metadata['book_cover']['url']
        else:
            return ""

    @property
    def page_cover(self) -> str:
        """Returns the page cover url or an empty string."""
        if 'page_cover' in self._metadata:
            return self._metadata['page_cover']['url']
        else:
            return ""

    @property
    def world(self) -> 'World':
        """The world this category is a part of."""
        if self._world is None:
            from pywaclient.models.world import World
            self._world = World(self._client, self._client.world.get(self._metadata['world']['id']))
        return self._world



#    Copyright 2020 Jonas Waeber
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from typing import Dict, Any


class Genre:

    def __init__(self, metadata: Dict[str, Any]):
        self._metadata = metadata

    @property
    def id(self) -> str:
        """Id of the genre."""
        return self._metadata['id']

    @property
    def name(self) -> str:
        """Name of the genre."""
        return self._metadata['name']
